#include "portability.h"
#include "playif.h"
#include "editif.h"
#include "volumeif.h"
#include "mp3_header.h"
#include "mpegindx.h"
#include "cursutil.h"
#include <sys/stat.h>

int curs_interactive_play_cb(void *data, long sec, long msec)
{
    /* found */
    int inch;
    play_cb_struct *ctx = (play_cb_struct *) data;
    mpegfio_iocallbacks *ttyio = ctx->ttyio;
    int  go = 1;
    long last_sec;
    long last_msec;
    int  volume;

    if (ttyio->getch &&
        (inch = ttyio->getch(NULL)) && inch != -1)
    {
        /*
         * Pause playback if <Enter> is pressed.
         * Resume if <Enter> is pressed again, otherwise stop
         */
        if (inch == '\n') {
            last_sec = mpgedit_play_current_sec(ctx->playctx);
            last_msec = mpgedit_play_current_msec(ctx->playctx);
            mpgedit_play_close(ctx->playctx);
            ctx->playctx = NULL;
            ttyio->printf(ttyio->ctx, 2, 0,
                "Playback paused (press <enter> to continue)\r");
            do {
                sleep(1);
            } while ((inch = ttyio->getch(NULL)) && inch == -1);

            ctx->playctx = mpgedit_play_init(ctx->filename);
            mpgedit_play_set_status_callback(ctx->playctx,
                                             curs_interactive_play_cb,
                                             ctx);
            mpgedit_play_seek_time(ctx->playctx, last_sec, last_msec);
        }
        
        if (inch == 'v'){
            volume = curs_volume_decrease(ctx->volctx);
            ttyio->printf(ttyio->ctx, 2, 74, "v:%3d", volume);
        }
        else if (inch == 'V'){
            volume = curs_volume_increase(ctx->volctx);
            ttyio->printf(ttyio->ctx, 2, 74, "v:%3d", volume);
        }
        else if (inch != '\n') {
            go = 0;
        }

        /*
         * Eat any remaining input characters. This happens when a multi-
         * character key, like an arrow key, is pressed.
         */
        do {
            inch = ttyio->getch(NULL);
        } while (inch && inch != -1);
    }
    if (sec > ctx->esec || (sec == ctx->esec && msec >= ctx->emsec)) {
       go = 0;
    }

    if (sec > ctx->tlast || !go) {
        ctx->tlast = sec;
        if (ttyio && ttyio->printf) {
            ttyio->printf(ttyio->ctx, 2, 0,
                "\rElapsed: %4d:%02d| %5d.%03ds|Frame: %-7d\r",
                   sec/60,
                   sec%60,
                   sec,
                   msec,
                   ctx->frame_count);
        }
    }
    ctx->frame_count++;
    return go;
}


int curs_interactive_play_file(char *filename,
                               char *stime_str,
                               char *dtime_str,
                               void *volctx,
                               mpegfio_iocallbacks *ttyio)
{
    int status = 0;
    mpeg_time stime; /* Start time */
    mpeg_time dtime; /* Play duration */
    long sec;
    long usec;
    long dsec;
    long dusec;
    int go;
    play_cb_struct   play_cb_data;

    memset(&play_cb_data, 0, sizeof(play_cb_data));
    mpeg_time_string2time(stime_str, &stime);
    mpeg_time_gettime(&stime, &sec,  &usec);

    mpeg_time_string2time(dtime_str, &dtime);
    mpeg_time_gettime(&dtime, &dsec, &dusec);


    play_cb_data.playctx = mpgedit_play_init(filename);
    mpgedit_play_seek_time(play_cb_data.playctx,
                           sec, usec/1000);
    play_cb_data.esec  = sec+dsec;
    play_cb_data.emsec = (usec+dusec)/1000;
    play_cb_data.ttyio = ttyio;
    play_cb_data.filename = filename;
    play_cb_data.volctx   = volctx;
    
    mpgedit_play_set_status_callback(play_cb_data.playctx,
                                     curs_interactive_play_cb, &play_cb_data);
    do {
        go = mpgedit_play_frame(play_cb_data.playctx);
    } while (go);
    mpgedit_play_close(play_cb_data.playctx);
    return status;
}


char *curs_validate_edit_times(editspec_t *edits)
{
    char *errstr = NULL;

    /*
     * Process -e and -f options.
     */
    if (!mpgedit_edit_times_init(edits)) {
        errstr = strdup("-f filename never specified");
#if 0
        fprintf(stderr, "%s: -f filename never specified\n", g_progname);
#endif
        return errstr;
    }

    if (mpgedit_editspec_get_length(edits) > 0) {
        errstr = mpgedit_edit_validate_times(edits, 0, -1);
    }
    else {
        errstr = strdup("no edits present");
    }
    return errstr;
}


int curs_build_indexfile(char *filename,
                         int (*callback)(void *, long, long, long, char *dummy),
                         void *callback_data)

{
    char *indx_filename;
    FILE *fp;
    struct stat sbuf;

    if (stat(filename, &sbuf) == -1) {
        return 1;
    }

    if (!S_ISREG(sbuf.st_mode)) {
        return 1;
    }

    fp = fopen(filename, "rb");
    if (!fp) {
        return 1;
    }

    indx_filename = build_index_filename(filename);
    if (!indx_filename) {
        return 1;
    }


    mp3header_index_create(filename, callback, callback_data, NULL, NULL);

    fclose(fp);
    free(indx_filename);
    return 0;
}


int curs_get_size_from_index(char *filename, long *sec, long *usec)
{
    FILE *indxfp;
    char *indx_filename;
    struct stat sbuf;
    long fsize = 0;

    if (stat(filename, &sbuf) == -1) {
        return 1;
    }

    if (!S_ISREG(sbuf.st_mode)) {
        return 1;
    }


    indx_filename = build_index_filename(filename);
    if (!indx_filename) {
        return 1;
    }
    indxfp = fopen(indx_filename, "r");
    if (!indxfp) {
        return 1;
    }

    fsize = mp3edit_get_size_from_index(indxfp, sec, usec);
    fclose(indxfp);
    free(indx_filename);
    return !(fsize > 0);
}


int curs_volume_decrease(void *volctx)
{
    int rvol;
    int lvol;
    int volume;

    if (!volctx) {
        return -1;
    }
 
    mpgedit_volume_get(volctx, &lvol, &rvol);
    volume = (lvol + rvol) / 2;

    volume -= 2;
    volume = (volume < 0) ? 0 : volume;
    mpgedit_volume_set(volctx, volume, volume);

    return volume;
}


int curs_volume_increase(void *volctx)
{
    int rvol;
    int lvol;
    int volume;

    if (!volctx) {
        return -1;
    }
 
    mpgedit_volume_get(volctx, &lvol, &rvol);
    volume = (lvol + rvol) / 2;

    volume += 2;
    volume = (volume > 100) ? 100 : volume;
    mpgedit_volume_set(volctx, volume, volume);

    return volume;
}
